#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <limits.h>
#include "IconEdit.h"
#include "IconEdit.rh"

#ifdef __BORLANDC__
#   define min(a,b) ((a)<=(b) ? (a) : (b))
#   define max(a,b) ((a)>=(b) ? (a) : (b))
#endif

HINSTANCE g_hinst = 0;
HACCEL    g_hacc  = 0;
HHOOK     g_hhook = 0;
HWND      g_hwnd  = 0;

void * __cdecl operator new(unsigned int n) {
    return (void *) GlobalAlloc(GPTR, n);
}

void __cdecl operator delete(void *p) { 
    GlobalFree(p);
}

int Width (const RECT& rc) { return rc.right  - rc.left; }
int Height(const RECT& rc) { return rc.bottom - rc.top ; }

struct BmpDC {
    BITMAP  bmi;
    HBITMAP hbmSave;
    HDC     hdc;

    void CreateDC(HBITMAP hbm) {
        GetObject(hbm, sizeof bmi, &bmi);
        hdc = CreateCompatibleDC(0);
        hbmSave = SelectBitmap(hdc, hbm);
    };

    BmpDC(HBITMAP hbm = 0) { // Constructor.
        if (hbm)
            CreateDC(hbm);
    }

    ~BmpDC() { // Destructor.
        if (hdc)
            DeleteDC(/* bDelete = */ TRUE);
    }

    void DeleteDC(BOOL bDelete = FALSE) {
        HBITMAP hbm = SelectBitmap(hdc, hbmSave);
        if (bDelete)
            DeleteBitmap(hbm);
        ::DeleteDC(hdc);
        hdc = 0;
    };

    void Blt(const BmpDC& bmpSrc, const DWORD dwROP = SRCCOPY) {
        BitBlt(hdc, 0, 0, min(bmi.bmWidth, bmpSrc.bmi.bmWidth),
            min(bmi.bmHeight, bmpSrc.bmi.bmHeight),
            bmpSrc.hdc, 0, 0, dwROP);
    }

    void Blt(HBITMAP hbm) {
        BmpDC bmp(hbm);
        Blt(bmp);
        bmp.DeleteDC();
    }

    HBITMAP Copy(const DWORD dwRop = SRCCOPY) const {
        HBITMAP hbmNew = CreateBitmap(bmi.bmWidth, bmi.bmHeight,
            bmi.bmPlanes, bmi.bmBitsPixel, 0);
        BmpDC bmpNew(hbmNew);
        bmpNew.Blt(*this, dwRop);
        bmpNew.DeleteDC();
        return hbmNew;
    }

    void Clear(COLORREF cr = RGB(0, 0, 0),
        int w = INT_MAX, int h = INT_MAX) 
    {
        HBRUSH hbr = CreateSolidBrush(cr);
        const RECT rc = { 0, 0,
            min(w, bmi.bmWidth), min(h, bmi.bmHeight) };
        FillRect(hdc, &rc, hbr);
        DeleteBrush(hbr);
    }

    void Clear(COLORREF cr, const BmpDC& bmpMask,
        BOOL bInvertMask = TRUE)
    {
        BmpDC bmpSrc(Copy());
        bmpSrc.Clear(cr);
        BmpDC bmpInvMask(bmpMask.Copy(NOTSRCCOPY));
        bmpSrc.Blt(bInvertMask ? bmpMask : bmpInvMask, SRCAND);
        Blt(bInvertMask ? bmpInvMask : bmpMask, SRCAND);
        Blt(bmpSrc, SRCPAINT);
    }

    void Flip(BOOL bHorz) {
        if (bHorz)
            StretchBlt(hdc, 0, 0, bmi.bmWidth, bmi.bmHeight,
               hdc, bmi.bmWidth - 1, 0, -bmi.bmWidth, bmi.bmHeight,
               SRCCOPY);
        else
            StretchBlt(hdc, 0, 0, bmi.bmWidth, bmi.bmHeight,
               hdc, 0, bmi.bmHeight - 1, bmi.bmWidth, -bmi.bmHeight,
               SRCCOPY);
    }

    void FloodFill(int x, int y, COLORREF cr) {
        HBRUSH hbrSave = SelectBrush(hdc, CreateSolidBrush(cr));
        const COLORREF crOrg = GetPixel(hdc, x, y);
        ExtFloodFill(hdc, x, y, crOrg, FLOODFILLSURFACE);
        DeleteBrush(SelectBrush(hdc, hbrSave));
    }

    COLORREF InvertColor(COLORREF cr) {
        const COLORREF crSave = GetPixel(hdc, 0, 0); // Save...
        SetPixel(hdc, 0, 0, cr);
        PatBlt(hdc, 0, 0, 1, 1, DSTINVERT);
        cr = GetPixel(hdc, 0, 0);
        SetPixel(hdc, 0, 0, crSave);          // ...and restore.
        return cr;
    }
};

struct Editor;
struct Tool { // Base class for all tools.
    enum MouseButtons { left = 0, right = 1, num_buttons = 2 };
    COLORREF crAnd, crXor;

    virtual void ButtonDown(Editor *, int x, int y, int nButton);
    virtual void MouseMove(Editor *, int, int) {}
};

struct Editor {
    ICON_DESCR *pid;
    Tool       *pTool;
    ICONINFO   ii;
    BmpDC      xor, and;
    HBITMAP    hbmSaveXor, hbmSaveAnd; // For undo.
    int        nCell, nIdx[Tool::num_buttons], nPalOffs;
    HWND      hwndZoom, hwndImg, hwndPal, hwndTB, hwndTB2, hwndNext;
    COLORREF   acrPal[2 + 16];
    BOOL       bDirty;

    void ScaleDown(int& x, int& y) const { x /= nCell; y /= nCell; }
    BOOL IsIcon() const { return IMAGE_ICON == pid->uiType; }

    Editor(ICON_DESCR *p, Tool *pFirstTool) { // Constructor.
        pid = p;
        if (IsIcon())
            GetIconInfo((HICON) pid->horg, &ii);
        else {
            BmpDC bmp((HBITMAP) pid->horg);
            ii.hbmColor = bmp.Copy();
            bmp.DeleteDC();
            GetObject(ii.hbmColor, sizeof xor.bmi, &xor.bmi);
            ii.hbmMask = CreateBitmap(xor.bmi.bmWidth,
                xor.bmi.bmHeight, 1, 1, 0);
            and.CreateDC(ii.hbmMask);
            and.Clear();
            and.DeleteDC();
        }
        pid->hnew = (HANDLE) CreateIconIndirect(&ii);
        and.CreateDC(ii.hbmMask );
        xor.CreateDC(ii.hbmColor);
        pTool = pFirstTool;
        const int nMax = max(xor.bmi.bmWidth, xor.bmi.bmHeight);
        nCell = max(min(8, 430 / nMax), 225 / xor.bmi.bmWidth);
        const COLORREF *pcr = (const COLORREF *)
            LockResource(LoadResource(g_hinst,
            FindResource(g_hinst, "Palette", RT_RCDATA)));
        for (int i = 0; i < 2 + 16; ++i)
            acrPal[i] = pcr[i];
        nPalOffs = IsIcon() ? 0 : 2;
        nIdx[Tool::left ] = 2; // Black.
        nIdx[Tool::right] = 3; // White.
    }

    void SetTool(Tool *pNewTool, int id) {
        delete pTool;
        pTool = pNewTool;
        SendMessage(hwndTB2, TB_CHECKBUTTON, id, TRUE);
    }

    void Paint(HDC hdc, int w, int h) const {
        HBRUSH hbr = CreateSolidBrush(acrPal[0]);
        DrawIconEx(hdc, 0, 0,
            (HICON) pid->hnew, w, h, 0, hbr, DI_NORMAL);
        DeleteBrush(hbr);
    }

    void Refresh(int x = -1, int y = -1) {
        and.DeleteDC();
        xor.DeleteDC();
        DestroyIcon((HICON) pid->hnew);
        pid->hnew = (HANDLE) CreateIconIndirect(&ii);
        and.CreateDC(ii.hbmMask );
        xor.CreateDC(ii.hbmColor);
        if (-1 == x)
            InvalidateRect(hwndZoom, 0, FALSE);
        else {
            RECT rc = { x      * nCell,  y      * nCell,
                       (x + 1) * nCell, (y + 1) * nCell };
            InvalidateRect(hwndZoom, &rc, FALSE);
        }
        UpdateWindow(hwndZoom);
        InvalidateRect(hwndImg, 0, FALSE);
        UpdateWindow(hwndImg);
        bDirty = TRUE;
    }

    void DeleteUndoInfo(void) {
        if (hbmSaveAnd)
            DeleteBitmap(hbmSaveAnd);
        if (hbmSaveXor)
            DeleteBitmap(hbmSaveXor);
    }

    void Remember(void) {
        DeleteUndoInfo();
        hbmSaveAnd = and.Copy();
        hbmSaveXor = xor.Copy();
        SendMessage(hwndTB, TB_ENABLEBUTTON, ID_UNDO, TRUE);
    }

    void DrawPalColor(HDC hdc, int left, int top, int nIdx) const {
        HBRUSH hbr =
            SelectBrush(hdc, CreateSolidBrush(acrPal[ nIdx ]));
        Rectangle(hdc, left + 1, top + 1, left + 15, top + 15);
        DeleteBrush(SelectBrush(hdc, hbr));
        if (nIdx < 2) { // It's a screen color.
            RECT rc = { left + 2, top + 2, left + 14, top + 14 };
            FrameRect(hdc, &rc, GetStockBrush(WHITE_BRUSH));
            InsetRect(&rc, 1, 1);
            FrameRect(hdc, &rc, GetStockBrush(BLACK_BRUSH));
        }
    }

    void SetScreenColor(int nIdx, COLORREF crBk) {
        acrPal[nIdx] = crBk;
        acrPal[1 - nIdx] = xor.InvertColor(acrPal[nIdx]);
        InvalidateRect(hwndPal , 0, FALSE);
        InvalidateRect(hwndZoom, 0, FALSE);
        InvalidateRect(hwndImg , 0, FALSE);
    }

    COLORREF GetXorColor(int nBtn) {
        const int nOffset = nIdx[nBtn] < 2 ? 2 : 0;
        return acrPal[nOffset + nIdx[nBtn]];
    }

    COLORREF GetAndColor(int nBtn) {
        return nIdx[nBtn] < 2 ? RGB(255, 255, 255) : RGB(0, 0, 0);
    }

    void OnUndo(void) {
        if (SendMessage(hwndTB, TB_ISBUTTONENABLED, ID_UNDO, 0)) {
            and.DeleteDC(/* bDelete = */ TRUE);
            xor.DeleteDC(/* bDelete = */ TRUE);
            DestroyIcon((HICON) pid->hnew);
            ii.hbmMask  = hbmSaveAnd;
            ii.hbmColor = hbmSaveXor;
            hbmSaveAnd = hbmSaveXor = 0;
            pid->hnew = (HANDLE) CreateIconIndirect(&ii);
            and.CreateDC(ii.hbmMask );
            xor.CreateDC(ii.hbmColor);
            InvalidateRect(hwndZoom, 0, FALSE);
            InvalidateRect(hwndImg , 0, FALSE);
            SendMessage(hwndTB, TB_ENABLEBUTTON, ID_UNDO, FALSE);
        }
    }

    void OnDelete(void) {
        Remember();
        and.Clear(GetAndColor(Tool::right));
        xor.Clear(GetXorColor(Tool::right));
        Refresh();
    }

    void OnCopyCut(HWND hwndDlg, int id) {
        if (OpenClipboard(hwndDlg)) {
            EmptyClipboard();
            HBITMAP hbmClp = xor.Copy();
            BmpDC bmpClp(hbmClp);
            Paint(bmpClp.hdc,
                bmpClp.bmi.bmWidth, bmpClp.bmi.bmHeight);
            bmpClp.DeleteDC();
            SetClipboardData(CF_BITMAP, hbmClp);
            CloseClipboard();
            if (ID_CUT == id)
                OnDelete();
        }
    }

    void OnPaste(HWND hwndDlg) {
        if (IsClipboardFormatAvailable(CF_BITMAP) &&
            OpenClipboard(hwndDlg))
        {
            Remember();
            and.Clear();
            BmpDC bmpClp((HBITMAP) GetClipboardData(CF_BITMAP));
            xor.Blt(bmpClp);
            bmpClp.DeleteDC();
            CloseClipboard();
            Refresh();
        }
    }

    void OnFlip(BOOL bHorz) {
        Remember();
        and.Flip(bHorz);
        xor.Flip(bHorz);
        Refresh();
    }

    void OnInvert(void) {
        Remember();
        xor.Blt(xor, DSTINVERT);
        Refresh();
    }

    void OnOKCancel(HWND hwndDlg, int id) {
        if (IDCANCEL == id && bDirty) {
            char szTitle[80];
            GetWindowText(hwndDlg, szTitle, sizeof szTitle);
            switch (MessageBox(hwndDlg,
                "Do you want to save your changes?",
                szTitle, MB_YESNOCANCEL | MB_ICONQUESTION)) {
            case IDCANCEL:
                return;
            case IDYES:
                id = IDOK;
                break;
            }
        }

        ChangeClipboardChain(hwndDlg, hwndNext);
        DeleteUndoInfo();
        if (IDOK != id || !IsIcon())
            DestroyIcon((HICON) pid->hnew);
        if (IDOK == id && !IsIcon())
            pid->hnew = xor.Copy();

        EndDialog(hwndDlg, id);
        delete pTool;
        delete this;
    }
};

void Tool::ButtonDown(Editor *p, int, int, int nButton) {
    crAnd = p->GetAndColor(nButton);
    crXor = p->GetXorColor(nButton);
}

struct PaintTool : public Tool {
    virtual void ButtonDown(Editor *p, int x, int y, int nButton) {
        Tool::ButtonDown(p, x, y, nButton);
        MouseMove(p, x, y);
    }

    virtual void MouseMove(Editor *p, int x, int y) {
        SetPixelV(p->xor.hdc, x, y, crXor);
        SetPixelV(p->and.hdc, x, y, crAnd);
        p->Refresh(x, y);
    }
};

struct EraseTool : public PaintTool {
    COLORREF crAndOld, crXorOld;

    virtual void ButtonDown(Editor *p, int x, int y, int nButton) {
        crAndOld = p->GetAndColor(1 - nButton);
        crXorOld = p->GetXorColor(1 - nButton);
        PaintTool::ButtonDown(p, x, y, nButton);
    }

    virtual void MouseMove(Editor *p, int x, int y) {
        if (GetPixel(p->xor.hdc, x, y) == crXorOld &&
            GetPixel(p->and.hdc, x, y) == crAndOld)
        {
            PaintTool::MouseMove(p, x, y);
        }
    }
};

struct FloodTool : public Tool {
    virtual void ButtonDown(Editor *p, int x, int y, int nButton) {
        const COLORREF crInverse = p->xor.InvertColor(
            GetPixel(p->xor.hdc, x, y));

        BmpDC bmpNew(p->xor.Copy());
        bmpNew.Clear(crInverse, p->and, 
            GetPixel(p->and.hdc, x, y) == RGB(0, 0, 0));

        BmpDC bmpChanged(bmpNew.Copy());
        bmpNew.FloodFill(x, y, crInverse);
        bmpChanged.Blt(bmpNew, SRCINVERT);

        // Set changed parts to new colors:
        p->xor.Clear(p->GetXorColor(nButton), bmpChanged);
        p->and.Clear(p->GetAndColor(nButton), bmpChanged);
        p->Refresh();
    }
};

struct RubberTool : public Tool {
    int xStart, yStart;

    RECT GetOperationRect(int x, int y) const {
        RECT rc = { min(xStart, x)    , min(yStart, y)    ,
                    max(xStart, x) + 1, max(yStart, y) + 1 };
        return rc;
    }

    virtual void ButtonDown(Editor *p, int x, int y, int nButton) {
        Tool::ButtonDown(p, x, y, nButton);
        xStart = x;
        yStart = y;
        MouseMove(p, x, y);
    }

    virtual void MouseMove(Editor *p, int x, int y) {
        p->and.Blt(p->hbmSaveAnd);
        p->xor.Blt(p->hbmSaveXor);
        RubberMouseMove(p, x, y);
        p->Refresh();
    }

    virtual void RubberMouseMove(Editor *, int, int) {}
};

struct LineTool : public RubberTool {
    virtual void RubberMouseMove(Editor *p, int x, int y) {
        HPEN hpenSave = SelectPen(p->xor.hdc,
            CreatePen(PS_SOLID, 1, crXor));
        POINT apt[] = { { xStart, yStart }, { x, y } };
        Polyline(p->xor.hdc, apt, 2);
        SetPixelV(p->xor.hdc, x, y, crXor);
        DeletePen(SelectPen(p->xor.hdc, hpenSave));

        hpenSave = SelectPen(p->and.hdc,
            CreatePen(PS_SOLID, 1, crAnd));
        Polyline(p->and.hdc, apt, 2);
        SetPixelV(p->and.hdc, x, y, crAnd);
        DeletePen(SelectPen(p->and.hdc, hpenSave));
    }
};

struct RectTool : public RubberTool {
    virtual void RubberMouseMove(Editor *p, int x, int y) {
        HPEN hpenSave = SelectPen(p->xor.hdc,
            CreatePen(PS_SOLID, 1, crXor));
        HBRUSH hbrSave = SelectBrush(p->xor.hdc,
            GetStockBrush(NULL_BRUSH));
        const RECT rc = GetOperationRect(x, y);
        Rectangle(p->xor.hdc, rc.left, rc.top, rc.right, rc.bottom);
        DeletePen(SelectPen(p->xor.hdc, hpenSave));
        SelectBrush(p->xor.hdc, hbrSave);

        hpenSave = SelectPen(p->and.hdc,
            CreatePen(PS_SOLID, 1, crAnd));
        hbrSave = SelectBrush(p->and.hdc,
            GetStockBrush(NULL_BRUSH));
        Rectangle(p->and.hdc, rc.left, rc.top, rc.right, rc.bottom);
        DeletePen(SelectPen(p->and.hdc, hpenSave));
        SelectBrush(p->and.hdc, hbrSave);
    }
};

struct FillRectTool : public RubberTool {
    virtual void RubberMouseMove(Editor *p, int x, int y) {
        const RECT rc = GetOperationRect(x, y);
        HBRUSH hbr = CreateSolidBrush(crXor);
        FillRect(p->xor.hdc, &rc, hbr);
        DeleteBrush(hbr);

        hbr = CreateSolidBrush(crAnd);
        FillRect(p->and.hdc, &rc, hbr);
        DeleteBrush(hbr);
    }
};

struct EllipseTool : public RubberTool {
    virtual void RubberMouseMove(Editor *p, int x, int y) {
        const RECT rc = GetOperationRect(x, y);
        HPEN hpenSave = SelectPen(p->xor.hdc,
            CreatePen(PS_SOLID, 1, crXor));
        HBRUSH hbrSave = SelectBrush(p->xor.hdc,
            GetStockBrush(NULL_BRUSH));
        Ellipse(p->xor.hdc, rc.left, rc.top, rc.right, rc.bottom);
        DeletePen(SelectPen(p->xor.hdc, hpenSave));
        SelectBrush(p->xor.hdc, hbrSave);

        hpenSave = SelectPen(p->and.hdc,
            CreatePen(PS_SOLID, 1, crAnd));
        hbrSave = SelectBrush(p->and.hdc,
            GetStockBrush(NULL_BRUSH));
        Ellipse(p->and.hdc, rc.left, rc.top, rc.right, rc.bottom);
        DeletePen(SelectPen(p->and.hdc, hpenSave));
        SelectBrush(p->and.hdc, hbrSave);
    }
};

struct FillEllipseTool : public RubberTool {
    virtual void RubberMouseMove(Editor *p, int x, int y) {
        const RECT rc = GetOperationRect(x, y);
        HPEN hpenSave = SelectPen(p->xor.hdc,
            CreatePen(PS_SOLID, 1, crXor));
        HBRUSH hbrSave = SelectBrush(p->xor.hdc,
            CreateSolidBrush(crXor));
        Ellipse(p->xor.hdc, rc.left, rc.top, rc.right, rc.bottom);
        DeletePen(SelectPen(p->xor.hdc, hpenSave));
        DeleteBrush(SelectBrush(p->xor.hdc, hbrSave));

        hpenSave = SelectPen(p->and.hdc,
            CreatePen(PS_SOLID, 1, crAnd));
        hbrSave = SelectBrush(p->and.hdc,
            CreateSolidBrush(crAnd));
        Ellipse(p->and.hdc, rc.left, rc.top, rc.right, rc.bottom);
        DeletePen(SelectPen(p->and.hdc, hpenSave));
        DeleteBrush(SelectBrush(p->and.hdc, hbrSave));
    }
};

Editor *GetEditor(HWND hwnd) {
    Editor *p = (Editor *) GetWindowLong(hwnd, GWL_USERDATA);
    if (0 == p)
        p = (Editor *) GetWindowLong(GetParent(hwnd), GWL_USERDATA);
    return p;
}

void ChangePalColor(HWND hwnd, BOOL fDbl, int x, int y, int n) {
    if (0 <= (y -= 32)) {
        Editor *p = GetEditor(hwnd);
        p->nIdx[n] = (y / 16) * 2 + (x / 16) % 2 + p->nPalOffs;
        InvalidateRect(hwnd, 0, FALSE);
        static COLORREF acrCustom[16] = { RGB(0, 0, 0) };
        CHOOSECOLOR cc = { sizeof(CHOOSECOLOR), hwnd, 0,
            p->acrPal[p->nIdx[n]], acrCustom, CC_RGBINIT };
        if (fDbl && p->nIdx[n] < 2 && ChooseColor(&cc))
            p->SetScreenColor(p->nIdx[n], cc.rgbResult);
    }
}

void Pal_OnPaint(HWND hwnd) {
    PAINTSTRUCT ps;
    BeginPaint(hwnd, &ps);

    const Editor *p = GetEditor(hwnd);
    p->DrawPalColor(ps.hdc, 5, 5, p->nIdx[Tool::left]);
    ExcludeClipRect(ps.hdc, 5, 5, 5 + 15, 5 + 15);
    p->DrawPalColor(ps.hdc, 11, 11, p->nIdx[Tool::right]);

    for (int nIdx = p->nPalOffs; nIdx < 18; ++nIdx) {
        const int left = ((nIdx - p->nPalOffs) % 2) * 16;
        const int top  = ((nIdx - p->nPalOffs) / 2) * 16 + 32;
        p->DrawPalColor(ps.hdc, left, top, nIdx);
    }

    EndPaint(hwnd, &ps);
}

void Pal_OnLButtonDown(HWND hwnd, BOOL fDbl, int x, int y, UINT) {
    ChangePalColor(hwnd, fDbl, x, y, 0);
}

void Pal_OnRButtonDown(HWND hwnd, BOOL fDbl, int x, int y, UINT) {
    ChangePalColor(hwnd, fDbl, x, y, 1);
}

LRESULT CALLBACK PalWndProc(
    HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
    HANDLE_MSG(hwnd, WM_PAINT        , Pal_OnPaint      );
    HANDLE_MSG(hwnd, WM_LBUTTONDOWN  , Pal_OnLButtonDown);
    HANDLE_MSG(hwnd, WM_RBUTTONDOWN  , Pal_OnRButtonDown);
    HANDLE_MSG(hwnd, WM_LBUTTONDBLCLK, Pal_OnLButtonDown);
    HANDLE_MSG(hwnd, WM_RBUTTONDBLCLK, Pal_OnRButtonDown);
    }

    return DefWindowProc(hwnd, msg, wParam, lParam);
}

BOOL Img_OnEraseBkgnd(HWND hwnd, HDC hdc) {
    const Editor *p = GetEditor(hwnd);
    if (4 < p->nCell && IDC_ZOOM == GetWindowID(hwnd)) {
        RECT rc;
        GetClientRect(hwnd, &rc);
        HBITMAP hbm = LoadBitmap(g_hinst, "Grid");
        HBRUSH hbr = CreatePatternBrush(hbm);
        FillRect(hdc, &rc, hbr);
        DeleteBrush(hbr);
        DeleteBitmap(hbm);
    }
    return TRUE;
}

void Img_OnPaint(HWND hwnd){
    PAINTSTRUCT ps;
    BeginPaint(hwnd, &ps);

    RECT rc;
    GetClientRect(hwnd, &rc);
    const Editor *p = GetEditor(hwnd);
    BOOL bGrid = 4 < p->nCell && IDC_ZOOM == GetWindowID(hwnd);
    if (bGrid) {
        for (int y = rc.top; y < rc.bottom; y += p->nCell)
            ExcludeClipRect(ps.hdc, 0, y, rc.right, y + 1);
        for (int x = rc.left; x < rc.right; x += p->nCell)
            ExcludeClipRect(ps.hdc, x, 0, x + 1, rc.bottom);
    }
    p->Paint(ps.hdc, rc.right - bGrid, rc.bottom - bGrid);
    EndPaint(hwnd, &ps);
}

static UINT g_nCurrButton;
static int g_xCurr, g_yCurr;

void Img_CommonButtonDown(HWND hwnd, int x, int y, UINT nButton) {
    if (0 == GetCapture()) {
        SetCapture(hwnd);
        Editor *p = GetEditor(hwnd);
        if (IDC_ZOOM == GetWindowID(hwnd))
            p->ScaleDown(x, y);
        p->Remember();
        g_nCurrButton = nButton;
        g_xCurr = x;
        g_yCurr = y;
        p->pTool->ButtonDown(p, x, y, nButton);
    }
}

void Img_OnLButtonDown(HWND hwnd, BOOL, int x, int y, UINT) {
    Img_CommonButtonDown(hwnd, x, y, Tool::left);
}

void Img_OnRButtonDown(HWND hwnd, BOOL, int x, int y, UINT) {
    Img_CommonButtonDown(hwnd, x, y, Tool::right);
}

void Img_OnMouseMove(HWND hwnd, int x, int y, UINT) {
    if (GetCapture() == hwnd) {
        Editor *p = GetEditor(hwnd);
        if (IDC_ZOOM == GetWindowID(hwnd))
            p->ScaleDown(x, y);
        if (g_xCurr != x || g_yCurr != y) {
            g_xCurr = x;
            g_yCurr = y;
            p->pTool->MouseMove(p, x, y);
        }
    }
}

void Img_OnCommonButtonUp(HWND hwnd, UINT nButton) {
    if (GetCapture() == hwnd && g_nCurrButton == nButton) {
        ReleaseCapture();
    }
}

void Img_OnLButtonUp(HWND hwnd, int, int, UINT) {
    Img_OnCommonButtonUp(hwnd, Tool::left);
}

void Img_OnRButtonUp(HWND hwnd, int, int, UINT) {
    Img_OnCommonButtonUp(hwnd, Tool::right);
}

LRESULT CALLBACK ImgWndProc(
    HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
    HANDLE_MSG(hwnd, WM_ERASEBKGND , Img_OnEraseBkgnd );
    HANDLE_MSG(hwnd, WM_PAINT      , Img_OnPaint      );
    HANDLE_MSG(hwnd, WM_LBUTTONDOWN, Img_OnLButtonDown);
    HANDLE_MSG(hwnd, WM_RBUTTONDOWN, Img_OnRButtonDown);
    HANDLE_MSG(hwnd, WM_MOUSEMOVE  , Img_OnMouseMove  );
    HANDLE_MSG(hwnd, WM_LBUTTONUP  , Img_OnLButtonUp  );
    HANDLE_MSG(hwnd, WM_RBUTTONUP  , Img_OnRButtonUp  );
    }

    return DefWindowProc(hwnd, msg, wParam, lParam);
}

#ifndef TBSTYLE_FLAT
#define TBSTYLE_FLAT    (0x0800)
#endif
#define WS_TOOLBAR \
    (TBSTYLE_FLAT | TBSTYLE_TOOLTIPS | CCS_TOP | CCS_NODIVIDER)

HWND CreateToolbar(HWND hwndParent) {
    HWND hwnd = CreateWindowEx(
        0, TOOLBARCLASSNAME, 0, WS_CHILD | WS_VISIBLE | WS_TOOLBAR,
        0, 0, 0, 0, hwndParent, (HMENU) ID_BAR, g_hinst, 0);
    SendMessage(hwnd, TB_BUTTONSTRUCTSIZE, sizeof(TBBUTTON), 0);

    TBADDBITMAP tbabmpStd = { HINST_COMMCTRL, IDB_STD_SMALL_COLOR };
    SendMessage(hwnd, TB_ADDBITMAP, 15, (LPARAM) &tbabmpStd);

    TBADDBITMAP tbabmp = { g_hinst, ID_BAR };
    SendMessage(hwnd, TB_ADDBITMAP, 5, (LPARAM) &tbabmp);

    static const TBBUTTON tb[] = {
        { 15 + 0    , IDOK     , TBSTATE_ENABLED, TBSTYLE_BUTTON },
        { 15 + 1    , IDCANCEL , TBSTATE_ENABLED, TBSTYLE_BUTTON },
        { 0         , 0        , 0              , TBSTYLE_SEP    },
        { STD_CUT   , ID_CUT   , TBSTATE_ENABLED, TBSTYLE_BUTTON },
        { STD_COPY  , ID_COPY  , TBSTATE_ENABLED, TBSTYLE_BUTTON },
        { STD_PASTE , ID_PASTE , TBSTATE_ENABLED, TBSTYLE_BUTTON },
        { STD_UNDO  , ID_UNDO  , 0                               },
        { 0         , 0        , 0              , TBSTYLE_SEP    },
        { STD_DELETE, ID_DELETE, TBSTATE_ENABLED                 },
        { 0         , 0        , 0              , TBSTYLE_SEP    },
        { 15 + 2    , ID_MIRROR, TBSTATE_ENABLED                 },
        { 15 + 3    , ID_FLIP  , TBSTATE_ENABLED                 },
        { 0         , 0        , 0              , TBSTYLE_SEP    },
        { 15 + 4    , ID_INVERT, TBSTATE_ENABLED                 },
    };

    SendMessage(hwnd, TB_ADDBUTTONS,
        sizeof(tb) / sizeof(tb[0]), (LPARAM)(TBBUTTON *) tb);

    return hwnd;
}

HWND CreateToolbar2(HWND hwndParent) {
    HWND hwnd = CreateWindowEx(
        0, TOOLBARCLASSNAME, 0, WS_CHILD | WS_VISIBLE | WS_TOOLBAR,
        0, 0, 0, 0, hwndParent, (HMENU) ID_BAR2, g_hinst, 0);
    SendMessage(hwnd, TB_BUTTONSTRUCTSIZE, sizeof(TBBUTTON), 0);

    TBADDBITMAP tbabmp = { g_hinst, ID_BAR2 };
    SendMessage(hwnd, TB_ADDBITMAP, 5, (LPARAM) &tbabmp);

    static const TBBUTTON tb[] = {
        { 0, ID_PAINT      , TBSTATE_CHECKED |
                             TBSTATE_ENABLED, TBSTYLE_CHECKGROUP },
        { 1, ID_ERASE      , TBSTATE_ENABLED, TBSTYLE_CHECKGROUP },
        { 2, ID_FILL       , TBSTATE_ENABLED, TBSTYLE_CHECKGROUP },
        { 3, ID_LINE       , TBSTATE_ENABLED, TBSTYLE_CHECKGROUP },
        { 4, ID_RECT       , TBSTATE_ENABLED, TBSTYLE_CHECKGROUP },
        { 5, ID_FILLRECT   , TBSTATE_ENABLED, TBSTYLE_CHECKGROUP },
        { 6, ID_ELLIPSE    , TBSTATE_ENABLED, TBSTYLE_CHECKGROUP },
        { 7, ID_FILLELLIPSE, TBSTATE_ENABLED, TBSTYLE_CHECKGROUP },
    };

    SendMessage(hwnd, TB_ADDBUTTONS,
        sizeof(tb) / sizeof(tb[0]), (LPARAM)(TBBUTTON *) tb);

    return hwnd;
}

BOOL Dlg_OnInitDialog(HWND hwnd, HWND, LPARAM lParam) {
    g_hwnd = hwnd;
    Editor *p = new Editor((ICON_DESCR *) lParam, new PaintTool);
    SetWindowLong(hwnd, GWL_USERDATA, (long) p);
    p->hwndZoom = GetDlgItem(hwnd, IDC_ZOOM);
    p->hwndImg  = GetDlgItem(hwnd, IDC_IMG);
    p->hwndPal  = GetDlgItem(hwnd, IDC_PAL);
    p->hwndTB   = CreateToolbar(hwnd);
    p->hwndTB2  = CreateToolbar2(hwnd);
    p->SetScreenColor(0, GetSysColor(COLOR_3DFACE));

    RECT rcTB, rcTB2, rcDlg, rcZoom, rcImg, rcPal;
    GetWindowRect(p->hwndTB, &rcTB);
    GetWindowRect(p->hwndTB2, &rcTB2);
    SetWindowPos(p->hwndTB2, 0, 0, Height(rcTB),
        Width(rcTB2), Height(rcTB2), SWP_NOZORDER);
    GetWindowRect(p->hwndTB2, &rcTB2);

    const int nToolbarHeights = rcTB2.bottom - rcTB.top;
    int nWid = p->nCell * p->xor.bmi.bmWidth  + (4 < p->nCell);
    int nHei = p->nCell * p->xor.bmi.bmHeight + (4 < p->nCell);
    SetWindowPos(p->hwndZoom, 0, 4, nToolbarHeights + 2,
        nWid + 2, nHei + 2, SWP_NOZORDER);
    SetWindowPos(p->hwndImg, 0, nWid + 10, nToolbarHeights + 2,
        p->and.bmi.bmWidth + 2, p->and.bmi.bmHeight + 2,
        SWP_NOZORDER);
    SetWindowPos(p->hwndPal, 0, nWid + 10,
        nToolbarHeights + p->and.bmi.bmHeight + 6 + 2,
        32 + 2, 162 + (p->IsIcon() ? 16 : 0), SWP_NOZORDER);

    GetWindowRect(p->hwndZoom, &rcZoom);
    MapWindowPoints(HWND_DESKTOP, hwnd, (POINT *) &rcZoom, 2);
    GetWindowRect(p->hwndImg, &rcImg);
    MapWindowPoints(HWND_DESKTOP, hwnd, (POINT *) &rcImg, 2);
    GetWindowRect(p->hwndPal, &rcPal);
    MapWindowPoints(HWND_DESKTOP, hwnd, (POINT *) &rcPal, 2);
    RECT rc = { 0, 0, max(rcImg.right, rcPal.right) + 4,
        max(rcZoom.bottom, rcPal.bottom) + 4 };
    AdjustWindowRectEx(&rc, GetWindowStyle(hwnd), FALSE,
        GetWindowExStyle(hwnd));
    GetWindowRect(hwnd, &rcDlg);
    SetWindowPos(hwnd, 0,
        rcDlg.left - (Width (rc) - Width (rcDlg)) / 2,
        rcDlg.top  - (Height(rc) - Height(rcDlg)) / 2,
        Width(rc), Height(rc), SWP_NOZORDER);

    p->hwndNext = SetClipboardViewer(hwnd);
    if (0 != p->pid->pszDlgTitle)
        SetWindowText(hwnd, p->pid->pszDlgTitle);
    SetFocus(hwnd);
    return FALSE;
}

void Dlg_OnCommand(HWND hwnd, int id, HWND, UINT) {
    Editor *p = GetEditor(hwnd);

    switch (id) {
    case IDCANCEL : case IDOK   : p->OnOKCancel(hwnd, id); break;
    case ID_DELETE:               p->OnDelete  (        ); break;
    case ID_CUT   : case ID_COPY: p->OnCopyCut (hwnd, id); break;
    case ID_PASTE :               p->OnPaste   (hwnd    ); break;
    case ID_MIRROR:               p->OnFlip    (TRUE    ); break;
    case ID_FLIP  :               p->OnFlip    (FALSE   ); break;
    case ID_INVERT:               p->OnInvert  (        ); break;
    case ID_UNDO  :               p->OnUndo    (        ); break;

    case ID_PAINT      : p->SetTool(new PaintTool      , id); break;
    case ID_FILL       : p->SetTool(new FloodTool      , id); break;
    case ID_LINE       : p->SetTool(new LineTool       , id); break;
    case ID_RECT       : p->SetTool(new RectTool       , id); break;
    case ID_FILLRECT   : p->SetTool(new FillRectTool   , id); break;
    case ID_ELLIPSE    : p->SetTool(new EllipseTool    , id); break;
    case ID_FILLELLIPSE: p->SetTool(new FillEllipseTool, id); break;
    case ID_ERASE      : p->SetTool(new EraseTool      , id); break;
    }
}

void Dlg_OnDrawClipboard(HWND hwnd) {
    const Editor *p = GetEditor(hwnd);
    if (IsWindow(p->hwndNext))
        FORWARD_WM_DRAWCLIPBOARD(p->hwndNext, SendMessage);
    SendMessage(p->hwndTB, TB_ENABLEBUTTON, ID_PASTE,
        IsClipboardFormatAvailable(CF_BITMAP));
}

void Dlg_OnChangeCBChain(HWND hwnd, HWND hwndRemove, HWND hwndNext)
{
    if (GetEditor(hwnd)->hwndNext == hwndRemove)
        GetEditor(hwnd)->hwndNext = hwndNext;
    else
        FORWARD_WM_CHANGECBCHAIN(GetEditor(hwnd)->hwndNext,
            hwndRemove, hwndNext, SendMessage);
}

BOOL CALLBACK DlgProc(
    HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
    HANDLE_MSG(hwnd, WM_INITDIALOG   , Dlg_OnInitDialog   );
    HANDLE_MSG(hwnd, WM_COMMAND      , Dlg_OnCommand      );
    HANDLE_MSG(hwnd, WM_DRAWCLIPBOARD, Dlg_OnDrawClipboard);
    HANDLE_MSG(hwnd, WM_CHANGECBCHAIN, Dlg_OnChangeCBChain);

    case WM_NOTIFY:
        TOOLTIPTEXT *pTTT = (TOOLTIPTEXT *) lParam;
        if (TTN_NEEDTEXT == pTTT->hdr.code) {
            pTTT->lpszText = MAKEINTRESOURCE(pTTT->hdr.idFrom);
            pTTT->hinst = g_hinst;
            return TRUE;
        }
    }

    return FALSE;
}

LRESULT CALLBACK HookProc(int nCode, WPARAM wParam, LPARAM lParam)
{
    if (MSGF_DIALOGBOX == nCode &&
        TranslateAccelerator(g_hwnd, g_hacc, (MSG *) lParam))
        return TRUE;
    return CallNextHookEx(g_hhook, nCode, wParam, lParam);
}

extern "C" int IEAPI EditIcon(ICON_DESCR *pid) {
    g_hhook = SetWindowsHookEx(
        WH_MSGFILTER, (HOOKPROC)HookProc, g_hinst,
        GetCurrentThreadId());
    const int nRet = DialogBoxParam(g_hinst, "IconEdit",
        pid->hwndParent, DlgProc, (LPARAM) pid);
    UnhookWindowsHookEx(g_hhook);
    return nRet;
}

extern "C" int APIENTRY _DllMainCRTStartup(
    HINSTANCE hinst, DWORD dwReason, LPVOID)
{
    if (DLL_PROCESS_ATTACH == dwReason) {
        g_hinst = hinst;
        g_hacc = LoadAccelerators(g_hinst, "AccTable");

        const WNDCLASS wcImg = { CS_GLOBALCLASS, ImgWndProc, 0, 0,
            g_hinst, 0, LoadCursor(0, IDC_ARROW), 0, 0, "Image" };
        RegisterClass(&wcImg);

        const WNDCLASS wcPal = { CS_GLOBALCLASS | CS_DBLCLKS,
            PalWndProc, 0, 0, g_hinst, 0, LoadCursor(0, IDC_ARROW),
            (HBRUSH)(COLOR_3DFACE + 1), 0, "Palette" };
        RegisterClass(&wcPal);
    }
    return TRUE;
}
