/*****
** MIDIRES.C : implementation file
** History   : 07-Jan-98 Davide Ficano (DaFi) Created
**             30-Jan-98 DaFi Added Win16 support
*****/

/* Functions must be exported */
#define EXPORT_DLL_DIRECTION

#define STRICT

#ifdef _MSC_VER
/* WINNT.H uses nameless structs, ... */
#pragma warning(disable: 4115 4201 4214)
#endif

#define WIN32_EXTRA_LEAN
#define WIN32_LEAN_AND_MEAN
#include "midires.h"

#ifdef _MSC_VER
#pragma warning(default: 4115 4201 4214)
#endif

static const TCHAR RT_MIDI[]   = _T("MIDI");
static const TCHAR szAppName[] = _T("MidiResPlayer");

HINSTANCE g_hInst  = NULL;
HWND g_hWndMidi    = NULL;
UINT g_mciDeviceID = 0;
TCHAR g_szFileName[MAX_PATH];

/*****
** PlayMidiSound
*****/

BOOL DLL_DIRECTION PlayMidiSound(LPCTSTR pszMidi,
                                 HINSTANCE hMidiInst,
                                 DWORD fdwSound)
{
    BOOL bRetVal = FALSE; /* pessimistic */
    BOOL bFromRes, bAsync;

    /* Check if in this system is installed 
    ** at least one midi output device */
    if (!midiOutGetNumDevs())
        return FALSE;

    /* Stop previous play (if any) */
    if (!pszMidi)
    {
        StopPlay();
        return TRUE;
    }

    /* Already playing. If another module is playing
    ** PlayFile detects condition and returns FALSE */
    if (g_hWndMidi)
        return FALSE;

    bFromRes = (fdwSound & SND_RESOURCE) != 0;

    /* Ambiguous:  do we play a file OR a resource ? */
    if (bFromRes == ((fdwSound & SND_FILENAME) != 0))
        return FALSE;

    /* If no sync flag is specified default playback is sync 
    ** If together are specified (foolish) default if async */
    bAsync = (fdwSound & SND_ASYNC) != 0;

    /* Simply used to receive MM_NOTIFY */
    if (!CreateHoodWnd())
        return FALSE;

    if (bFromRes)
        bRetVal = PlayResource(hMidiInst, pszMidi, bAsync);
    else
        bRetVal = PlayFile(pszMidi, bAsync) == 0;

    /* If play in sync then window must be destroyed */
    if (!bRetVal || !bAsync)
        DestroyWindow(g_hWndMidi);

    return bRetVal;
}

/*****
** Support functions
*****/

void StopPlay()
{
    if (g_mciDeviceID)
        mciSendCommand(g_mciDeviceID, MCI_STOP, MCI_WAIT, 0);

    RemoveFile();

    if (g_hWndMidi)
    {
        DestroyWindow(g_hWndMidi);

        g_hWndMidi = NULL;
    }
}

BOOL InitInstance(HINSTANCE hInst)
{
    WNDCLASS wc;

    g_hInst = hInst;

    wc.hCursor       = NULL;
    wc.hIcon         = NULL;
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = szAppName;
    wc.hbrBackground = NULL;
    wc.hInstance     = g_hInst;
    wc.style         = 0;
    wc.lpfnWndProc   = (WNDPROC)HoodWndProc;
    wc.cbWndExtra    = 0;
    wc.cbClsExtra    = 0;
    
    return (RegisterClass(&wc) != 0);
}

BOOL CreateHoodWnd()
{
    g_hWndMidi = CreateWindow(szAppName, szAppName, WS_POPUP,
                              0, 0, 0, 0,
                              NULL, NULL, g_hInst, NULL);
    
    return g_hWndMidi != NULL;
}

LRESULT CALLBACK HoodWndProc(HWND hwnd, UINT msg, 
                             WPARAM wParam, LPARAM lParam)
{
    switch (msg) 
    {
        HANDLE_MSG(hwnd, WM_NCDESTROY, OnNcDestroy);
        HANDLE_MSG(hwnd, MM_MCINOTIFY, OnMCINotify);
    }

    return DefWindowProc(hwnd, msg, wParam, lParam);
}

/***** 
** When window is destroyed close device (if opened)
** remove file (if exists)
*****/

void OnNcDestroy(HWND hWnd) 
{
    UNREFERENCED_PARAMETER(hWnd);

    RemoveFile();

    g_hWndMidi = NULL;
}

void OnMCINotify(HWND hWnd, UINT deviceID, UINT message)
{
    UNREFERENCED_PARAMETER(hWnd);
    UNREFERENCED_PARAMETER(deviceID);
    UNREFERENCED_PARAMETER(message);

    DestroyWindow(g_hWndMidi);
}

BOOL PlayResource(HINSTANCE hMidiInst, 
                  LPCTSTR pszResource, BOOL bAsync)
{
    if (!TransferResource(hMidiInst, pszResource))
        return FALSE;

    /* PlayFile returns error code, 0 means ALL OK */
    return PlayFile(g_szFileName, bAsync) == 0;
}

BOOL TransferResource(HINSTANCE hInst, LPCTSTR pszResName)
{
    HRSRC hRes      = NULL;
    LPBYTE pRes     = NULL;
    HGLOBAL hResMem = NULL;
    BOOL bRetVal    = FALSE;

    /* If hInst == NULL searches in module associated with the
    ** image file used by the OS to create the current process */
    hRes = FindResource(hInst, pszResName, RT_MIDI);

    /* Try to write temp MIDI file */
    if (hRes && (hResMem = LoadResource(hInst, hRes)) != NULL)
        if ((pRes = (LPBYTE)LockResource(hResMem)) != NULL)
            bRetVal = WriteMidi(pRes, SizeofResource(hInst, hRes));

    if (pRes)
        UnlockResource(hResMem);
    if (hResMem)
        FreeResource(hResMem);

    return bRetVal;
}

/*****
** Determine a temp file name then write midi buffer to it
*****/
#ifdef WIN32
BOOL WriteMidi(LPBYTE pBuf, DWORD dwSize)
{
    TCHAR szTempPath[MAX_PATH];
    HANDLE hFile;
    DWORD dwStored;
    BOOL bRetVal;

    if (!pBuf || !*pBuf || dwSize <= 0)
        return FALSE;

    /* Get a temp file name */
    GetTempPath(MAX_PATH, szTempPath);
    if (!GetTempFileName(szTempPath, _T("mid"), 0, g_szFileName))
        return FALSE;

    hFile = CreateFile(g_szFileName, GENERIC_WRITE, 0,
                       NULL, CREATE_ALWAYS, 
                       FILE_ATTRIBUTE_TEMPORARY, NULL);

    if (hFile == INVALID_HANDLE_VALUE)
    {
        g_szFileName[0] = _T('\0');
        return FALSE;
    }

    bRetVal = WriteFile(hFile, pBuf, dwSize, &dwStored, NULL);

    CloseHandle(hFile);

    if (!bRetVal || dwStored != dwSize)
        RemoveFile();

    return bRetVal;
}
#else
/* WriteMidi for Win16 */
BOOL WriteMidi(LPBYTE pBuf, DWORD dwSize)
{
    BYTE TempDrive;
    HFILE hFile;
    long ByteStored;
    BOOL bRetVal;

    if (!pBuf || !*pBuf || dwSize <= 0)
        return FALSE;

    /* Get a temp file name */
    TempDrive = GetTempDrive(0);
    if (!GetTempFileName(TempDrive, _T("mid"), 0, g_szFileName))
        return FALSE;

    hFile = _lopen(g_szFileName, WRITE);

    if (hFile == HFILE_ERROR)
    {
        g_szFileName[0] = _T('\0');
        return FALSE;
    }

    ByteStored = _hwrite(hFile, pBuf, dwSize);
    bRetVal = ByteStored == (long)dwSize;

    _lclose(hFile);

    if (!bRetVal)
        RemoveFile();

    return bRetVal;
}
#endif /* WIN32 */

/*****
** This function is adapted from a Books on line (BOL) sample
** Search "Playing a MIDI File" in BOL
*****/

DWORD PlayFile(LPCTSTR lpszMIDIFileName, BOOL bAsync)
{
    DWORD dwReturn;
    MCI_OPEN_PARMS mciOpenParms;
    MCI_PLAY_PARMS mciPlayParms;

    /* Open the device by specifying the device and filename.
    ** MCI will attempt to choose the MIDI mapper
    ** as the output port.*/
    mciOpenParms.lpstrDeviceType = _T("sequencer");
    mciOpenParms.lpstrElementName = lpszMIDIFileName;

    g_mciDeviceID = 0;

    dwReturn = mciSendCommand((UINT)NULL, MCI_OPEN,
                              MCI_OPEN_TYPE | MCI_OPEN_ELEMENT,
                              (DWORD)(LPVOID) &mciOpenParms);
    if (dwReturn)
    {
        /* Failed to open device. Don't close it;
        ** just return error.*/
        return (dwReturn);
    }

    /* The device opened successfully; get the device ID.*/
    g_mciDeviceID = mciOpenParms.wDeviceID;

    /* Begin playback. The window procedure function for
    ** the parent window will be notified with an MM_MCINOTIFY
    ** message when playback is complete.
    ** At this time, the window procedure closes the device.*/
    mciPlayParms.dwCallback = (DWORD)g_hWndMidi;

    dwReturn = mciSendCommand(g_mciDeviceID, MCI_PLAY,
                          MCI_NOTIFY | (bAsync ? 0 : MCI_WAIT),
                          (DWORD)(LPVOID) &mciPlayParms);
    if (dwReturn)
    {
        mciSendCommand(g_mciDeviceID, MCI_CLOSE, MCI_WAIT, 0);
        g_mciDeviceID = 0;

        return (dwReturn);
    }

    return 0;
}

void RemoveFile()
{
    if (g_mciDeviceID)
    {
        mciSendCommand(g_mciDeviceID, MCI_CLOSE, MCI_WAIT, 0);
        g_mciDeviceID = 0;
    }

    if (g_szFileName[0])
    {
#ifdef WIN32
        DeleteFile(g_szFileName);
#else
        OFSTRUCT of;

        OpenFile(g_szFileName, &of, OF_DELETE);
#endif
        g_szFileName[0] = _T('\0');
    }
}

#ifdef WIN32
BOOL WINAPI DllMain(HINSTANCE hInst, DWORD dwReason, LPVOID lpNot)
{
    UNREFERENCED_PARAMETER(lpNot);

    if (dwReason == DLL_PROCESS_ATTACH)
    {
        g_szFileName[0] = _T('\0');
        if (!InitInstance(hInst))
            return FALSE;
    }

    return TRUE;
}
#else
int CALLBACK LibMain(HINSTANCE hInstance, WORD wDataSeg,
                     WORD cbHeapSize, LPSTR lpCmdLine)
{
    UNREFERENCED_PARAMETER(wDataSeg);
    UNREFERENCED_PARAMETER(cbHeapSize);
    UNREFERENCED_PARAMETER(lpCmdLine);

    if (!g_hInst)
    {
        g_szFileName[0] = _T('\0');
        if (!InitInstance(hInstance))
            return 0;
    }
    return 1;
}
#endif
